"""
Authentication backend
"""
import json
from typing import Optional, List, Dict

from .base import HTTPBase, ResponseBase
from .exceptions import InvalidPath, InternalServerError


class LDAPAuthBackend(HTTPBase):
    """
    Vault LDAP authentication backend
    """
    def __init__(self, *args, mount_path: str = 'ldap', **kwargs) -> None:
        super(LDAPAuthBackend, self).__init__(*args, **kwargs)

        self.mount_path = 'auth/' + mount_path

    async def mount(self, mount_path: str,
                    url: str, binddn: str = '', bindpass: str = '', userdn: str = '', userattr: str = 'uid', discoverdn: bool = False, deny_null_bind: bool = True, upndomain: str = '',
                    groupfilter: str = '(|(memberUid={{.Username}})(member={{.UserDN}})(uniqueMember={{.UserDN}}))', groupdn: str = '', groupattr: str = 'cn',
                    description: str = '', starttls: bool = False, tls_min_version: str = 'tls12', tls_max_version: str = 'tls12', insecure_tls: bool = False, certificate: str = ''):
        """
        Mount an LDAP backend

        :param mount_path: Mount path
        :param url: LDAP URL
        :param binddn: Bind DN
        :param bindpass: Bind password
        :param userdn: User DN
        :param userattr: User attribute
        :param discoverdn: If true use an anonymous bind to find user DN
        :param deny_null_bind: Deny empty passwords
        :param upndomain: userPrincipalDomain for Active Directory
        :param groupfilter: Group filter
        :param groupdn: Group DN
        :param groupattr: Group attribute
        :param description: Mount description
        :param starttls: Use StartTLS
        :param tls_min_version: TLS minimum version
        :param tls_max_version: TLS maximum version
        :param insecure_tls: Skip SSL cert verification
        :param certificate: CA certificate
        """
        # pylint: disable=too-many-arguments,too-many-locals
        payload = {
            'type': 'ldap',
            'description': description,
        }

        payload2 = {
            'url': url,
            'binddn': binddn,
            'bindpass': bindpass,
            'userdn': userdn,
            'userattr': userattr,
            'discoverdn': discoverdn,
            'deny_null_bind': deny_null_bind,
            'upndomain': upndomain,

            'groupfilter': groupfilter,
            'groupdn': groupdn,
            'groupattr': groupattr,

            'starttls': starttls,
            'tls_min_version': tls_min_version,
            'tls_max_version': tls_max_version,
            'insecure_tls': insecure_tls,
            'certificate': certificate
        }

        await self._post(['sys/auth', mount_path], payload=payload)
        await self._post(['auth', mount_path, 'config'], payload=payload2)

    async def get_config(self, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        Get LDAP config
        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._get([self.mount_path, 'config'], wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def list_groups(self, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        List LDAP groups

        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._list([self.mount_path, 'groups'], wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def create_group(self, name: str, policies: List[str]):
        """
        Create group (groups are local to vault but will be matched to groups coming from LDAP)

        :param name: Group name
        :param policies: Policies
        """
        payload = {'policies': ','.join(policies)}
        await self._post([self.mount_path, 'groups', name], payload=payload)

    async def read_group(self, name: str, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        Read group info

        :param name: Group name
        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._get([self.mount_path, 'groups', name], wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def delete_group(self, name: str):
        """
        Delete group

        :param name: Group name
        """
        await self._delete([self.mount_path, 'groups', name])

    async def list_users(self, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        List users

        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._list([self.mount_path, 'users'], wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def create_user(self, name: str, policies: List[str]):
        """
        Create user
        :param name: Username
        :param policies: policies
        """
        payload = {'policies': ','.join(policies)}
        await self._post([self.mount_path, 'users', name], payload=payload)

    async def read_user(self, name: str, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        Read user config

        :param name: Username
        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._get([self.mount_path, 'users', name], wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def delete_user(self, name: str):
        """
        Delete user

        :param name: Username
        """
        await self._delete([self.mount_path, 'users', name])

    async def login(self, username: str, password: str, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        Authorise against LDAP

        :param username: Username
        :param password: Password
        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._post([self.mount_path, 'login', username], payload={'password': password}, wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)


class GitHubAuthBackend(HTTPBase):
    """
    Vault GitHub authentication backend
    """
    def __init__(self, *args, mount_path: str = 'github', **kwargs) -> None:
        super(GitHubAuthBackend, self).__init__(*args, **kwargs)

        self.mount_path = 'auth/' + mount_path

    async def mount(self, mount_path: str, organization: str, description: str = '', base_url: Optional[str] = None, ttl: Optional[str] = None, max_ttl: Optional[str] = None):
        """
        Mount the GitHub backend againt the given path

        :param mount_path: Mount path
        :param organization: GitHub Organisation
        :param description: Description
        :param base_url: Base URL
        :param ttl: TTL
        :param max_ttl: Max TTL
        """
        payload = {
            'type': 'github',
            'description': description,
        }

        payload2 = {
            'organization': organization
        }
        if base_url is not None:
            payload2['base_url'] = base_url
        if ttl is not None:
            payload2['ttl'] = ttl
        if max_ttl is not None:
            payload2['max_ttl'] = max_ttl

        await self._post(['sys/auth', mount_path], payload=payload)
        await self._post(['auth', mount_path, 'config'], payload=payload2)

    async def map_team_policy(self, team_name: str, policies: List[str]):
        """
        Map a policy to a team

        :param team_name: Team name
        :param policies: Policies
        """
        team_name = team_name.lower().replace(' ', '-')

        payload = {'value': ','.join(policies)}

        await self._post([self.mount_path, 'map/teams', team_name], payload=payload)

    async def list_teams(self, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        List Teams

        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._list([self.mount_path, 'map/teams'], wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def get_team(self, team_name: str, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        Get team info

        :param team_name: Team name
        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        team_name = team_name.lower().replace(' ', '-')

        response = await self._get([self.mount_path, 'map/teams', team_name], wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def del_team(self, team_name: str):
        """
        Delete team

        :param team_name: Team name
        """
        team_name = team_name.lower().replace(' ', '-')

        await self._delete([self.mount_path, 'map/teams', team_name])

    async def map_user_policy(self, username: str, policies: List[str]):
        """
        Map policy to user

        :param username: Username
        :param policies: Policies
        """
        payload = {'value': ','.join(policies)}

        await self._post([self.mount_path, 'map/users', username], payload=payload)

    async def list_users(self, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        List users

        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._list([self.mount_path, 'map/users'], wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def get_user(self, user_name: str, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        Get user info

        :param user_name: Username
        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        user_name = user_name.lower().replace(' ', '-')

        response = await self._get([self.mount_path, 'map/users', user_name], wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def del_user(self, user_name: str):
        """
        Delete user

        :param user_name: Username
        """
        user_name = user_name.lower().replace(' ', '-')

        await self._delete([self.mount_path, 'map/users', user_name])

    async def login(self, token: str, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        GitHub token authentication

        :param token: Token
        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._post([self.mount_path, 'login'], payload={'token': token}, wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)


class UserPassAuthBackend(HTTPBase):
    """
    Vault User/Password authentication backend
    """
    def __init__(self, *args, mount_path: str = 'userpass', **kwargs) -> None:
        super(UserPassAuthBackend, self).__init__(*args, **kwargs)

        self.mount_path = 'auth/' + mount_path

    async def mount(self, mount_path: str, description: str = ''):
        """
        Mount the user/password backend againt the given path

        :param mount_path: Mount path
        :param description: Description
        """
        payload = {
            'type': 'userpass',
            'description': description
        }
        await self._post(['sys/auth', mount_path], payload=payload)

    async def read(self, username: str, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        Get user

        :param username: Username
        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._get([self.mount_path, 'users', username], wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def create(self, username: str, password: str, policies: Optional[List[str]] = None, ttl: Optional[int] = None, max_ttl: Optional[int] = None):
        """
        Create user

        :param username: Username
        :param password: Password
        :param policies: Policies
        :param ttl: TTL
        :param max_ttl: Max TTL
        """
        payload = {
            'password': password,
        }
        if policies is not None:
            payload['policies'] = ','.join(policies)

        if ttl is not None:
            payload['ttl'] = ttl
        if max_ttl is not None:
            payload['max_ttl'] = max_ttl

        await self._post([self.mount_path, 'users', username], payload=payload)

    async def update(self, username: str, policies: Optional[List[str]] = None, ttl: Optional[int] = None, max_ttl: Optional[int] = None):
        """
        Update user

        :param username: Username
        :param policies: Policies
        :param ttl: TTL
        :param max_ttl: Max TTL
        """
        if policies is not None or ttl is not None or max_ttl is not None:
            payload = {}
            if policies is not None:
                payload['policies'] = ','.join(policies)
            if ttl is not None:
                payload['ttl'] = ttl
            if max_ttl is not None:
                payload['max_ttl'] = max_ttl

            await self._post([self.mount_path, 'users', username], payload=payload)

    async def delete(self, username: str):
        """
        Delete user

        :param username: Username
        """
        await self._delete([self.mount_path, 'users', username])

    async def update_password(self, username: str, password: str):
        """
        Update user password

        :param username: Username
        :param password: Password
        """
        await self._post([self.mount_path, 'users', username, 'password'], payload={'password': password})

    async def update_policies(self, username: str, policies: List[str]):
        """
        Update policies associated to a user

        :param username: Username
        :param policies: Policies
        """
        await self._post([self.mount_path, 'users', username, 'policies'], payload={'policies': ','.join(policies)})

    async def login(self, username: str, password: str, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        Authenticate with username and password

        :param username: Username
        :param password: Password
        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._post([self.mount_path, 'login', username], payload={'password': password}, wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def list(self, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        List users

        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._list([self.mount_path, 'users'], wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)


class RadiusAuthBackend(HTTPBase):
    """
    Vault RADIUS authentication backend
    """
    def __init__(self, *args, mount_path: str = 'radius', **kwargs) -> None:
        super(RadiusAuthBackend, self).__init__(*args, **kwargs)

        self.mount_path = 'auth/' + mount_path

    async def mount(self, mount_path: str, radius_host: str, secret: str, radius_port: int = 1812, unregistered_user_policies: Optional[List[str]] = None, dial_timeout: int = 10,
                    read_timeout: int = 10, nas_port: int = 10, description: str = ''):
        """
        Mount RADIUS authentication backend against the given path

        :param mount_path: Mount path
        :param radius_host: RADIUS host
        :param secret: RADIUS shared secret
        :param radius_port: RADIUS port
        :param unregistered_user_policies: Default policies for unregistered users
        :param dial_timeout: RADIUS dial timeout
        :param read_timeout: RADIUS read timeout
        :param nas_port: NAD port number
        :param description: Mount description
        """
        payload = {
            'type': 'radius',
            'description': description,
        }

        payload2 = {
            'host': radius_host,
            'port': radius_port,
            'secret': secret,
            'dial_timeout': dial_timeout,
            'read_timeout': read_timeout,
            'nas_port': nas_port

        }
        if unregistered_user_policies is not None:
            payload2['unregistered_user_policies'] = ','.join(unregistered_user_policies)

        await self._post(['sys/auth', mount_path], payload=payload)
        await self._post(['auth', mount_path, 'config'], payload=payload2)

    async def create(self, username: str, policies: Optional[List[str]] = None):
        """
        Create a user

        :param username: Username
        :param policies: Policies
        """
        payload = {}
        if policies is not None:
            payload['policies'] = ','.join(policies)

        await self._post([self.mount_path, 'users', username], payload=payload)

    async def update(self, username: str, policies: Optional[List[str]] = None):
        """
        Update policies associated to a user

        :param username: Username
        :param policies: Policies
        """
        await self.create(username, policies)

    async def read(self, username: str, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        Get user

        :param username: Username
        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._get([self.mount_path, 'users', username], wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def list(self, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        List users

        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._list([self.mount_path, 'users'], wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def delete(self, username: str):
        """
        Delete user

        :param username: Username
        """
        await self._delete([self.mount_path, 'users', username])

    async def login(self, username: str, password: str, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        Authenticate with RADIUS using username and password

        :param username: Username
        :param password: Password
        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        payload = {
            'username': username,
            'password': password
        }

        response = await self._post([self.mount_path, 'login'], payload=payload, wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)


class ApproleBackend(HTTPBase):
    """
    Vault AppRole authentication backend
    """
    def __init__(self, *args, mount_path: str = 'approle', **kwargs) -> None:
        super(ApproleBackend, self).__init__(*args, **kwargs)

        self.mount_path = 'auth/' + mount_path

    async def mount(self, mount_path: str, description: str = ''):
        """
        Mount AppRole backed against the given path

        :param mount_path: Mount path
        :param description: Mount description
        """
        payload = {
            'type': 'approle',
            'description': description,
        }

        await self._post(['sys/auth', mount_path], payload=payload)

    async def list(self, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        List roles

        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._list([self.mount_path, 'role'], wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def create(self, name: str, bind_secret_id: bool = True, bound_cidr_list: Optional[List[str]] = None, policies: Optional[List[str]] = None,
                     secret_id_num_uses: int = 0, secret_id_ttl: int = 0, token_num_uses: int = 0, token_ttl: int = 0, token_max_ttl: int = 0, period: int = 0):
        """
        Create AppRole

        :param name: Role Name
        :param bind_secret_id: Require secret_id to be presented when logging in using this AppRole
        :param bound_cidr_list: List of CIDR blocks
        :param policies: List of policies
        :param secret_id_num_uses: Number of times any particular SecretID can be used to fetch a token from this AppRole
        :param secret_id_ttl: Secret ID TTL
        :param token_num_uses: Number of times issued tokens can be used
        :param token_ttl: Token TTL
        :param token_max_ttl: Token Max TTL
        :param period: Periodic token TTL
        """
        payload = {
            'bind_secret_id': bind_secret_id,
            'secret_id_num_uses': secret_id_num_uses,
            'secret_id_ttl': secret_id_ttl,
            'token_num_uses': token_num_uses,
            'token_ttl': token_ttl,
            'token_max_ttl': token_max_ttl,
            'period': period
        }
        if bound_cidr_list is not None:
            payload['bound_cidr_list'] = ','.join(bound_cidr_list)
        if policies is not None:
            payload['policies'] = ','.join(policies)

        await self._post([self.mount_path, 'role', name], payload=payload)

    async def update(self, name: str, bind_secret_id: bool = True, bound_cidr_list: Optional[List[str]] = None, policies: Optional[List[str]] = None,
                     secret_id_num_uses: int = 0, secret_id_ttl: int = 0, token_num_uses: int = 0, token_ttl: int = 0, token_max_ttl: int = 0, period: int = 0):
        """
        Update AppRole

        :param name: Role Name
        :param bind_secret_id: Require secret_id to be presented when logging in using this AppRole
        :param bound_cidr_list: List of CIDR blocks
        :param policies: List of policies
        :param secret_id_num_uses: Number of times any particular SecretID can be used to fetch a token from this AppRole
        :param secret_id_ttl: Secret ID TTL
        :param token_num_uses: Number of times issued tokens can be used
        :param token_ttl: Token TTL
        :param token_max_ttl: Token Max TTL
        :param period: Periodic token TTL
        """
        await self.create(name, bind_secret_id, bound_cidr_list, policies, secret_id_num_uses, secret_id_ttl, token_num_uses, token_ttl, token_max_ttl, period)

    async def read(self, name: str, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        Get Role

        :param name: Role name
        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._get([self.mount_path, 'role', name], wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def delete(self, name: str):
        """
        Delete role

        :param name: Role name
        """
        await self._delete([self.mount_path, 'role', name])

    async def read_roleid(self, name: str, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        Get role ID

        :param name: Role name
        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._get([self.mount_path, 'role', name, 'role-id'], wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def set_roleid(self, name: str, role_id: str):
        """
        Set the role ID of a given role, overwriting what it was before

        :param name: Role name
        :param role_id: Role ID
        """
        await self._post([self.mount_path, 'role', name, 'role-id'], payload={'role_id': role_id})

    async def generate_secret_id(self, name: str, cidr_list: Optional[List[str]] = None, metadata: Optional[dict] = None, wrap_ttl: Optional[int] = None, custom_id: str = None) -> ResponseBase:
        """
        Generate random secret ID

        :param name: Role Name
        :param cidr_list: Optional list of CIDRs
        :param metadata: Secret ID metadata
        :param wrap_ttl: Wrap TTL
        :param custom_id: Custom secret ID instead of a randomly generated one
        :return: Response
        """
        payload = {

        }
        url_part = 'secret-id'

        if cidr_list is not None:
            payload['cidr_list'] = ','.join(cidr_list)
        if metadata is not None:
            payload['metadata'] = json.dumps(metadata)

        if custom_id is not None:
            url_part = 'custom-secret-id'
            payload['secret_id'] = custom_id

        response = await self._post([self.mount_path, 'role', name, url_part], payload=payload, wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def list_secret_ids(self, name: str, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        List secret IDs for a given role

        :param name: Role name
        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._list([self.mount_path, 'role', name, 'secret-id'], wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def lookup_secret_id(self, name: str, secret_id: Optional[str] = None, secret_id_accessor: Optional[str] = None, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        Lookup Secret ID

        :param name: Role name
        :param secret_id: Secret ID
        :param secret_id_accessor: Secret ID accessor
        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        if secret_id is None and secret_id_accessor is None:
            raise ValueError('secret_id or secret_id_accessor must be provided')

        try:
            if secret_id is not None:
                response = await self._post([self.mount_path, 'role', name, 'secret-id/lookup'], payload={'secret_id': secret_id}, wrap_ttl=wrap_ttl)
            else:
                response = await self._post([self.mount_path, 'role', name, 'secret-id-accessor/lookup'], payload={'secret_id_accessor': secret_id_accessor}, wrap_ttl=wrap_ttl)
        except InternalServerError as err:  # Catch if accessor not found and convert into invalid path, theres probably a better way, its late, im tired.
            if 'failed to find accessor entry for secret_id' in str(err):
                raise InvalidPath("Secret does not exist")
            raise
        else:
            if response.status == 204:
                raise InvalidPath("Secret does not exist")
            else:
                json_data = await response.json()
                return ResponseBase(json_dict=json_data, request_func=self._request)

    async def destroy_secret_id(self, name: str, secret_id: Optional[str] = None, secret_id_accessor: Optional[str] = None):
        """
        Delete secret ID

        :param name: Role Name
        :param secret_id: Secret ID
        :param secret_id_accessor: Secret ID accessor
        :return:
        """
        if secret_id is None and secret_id_accessor is None:
            raise ValueError('secret_id or secret_id_accessor must be provided')

        if secret_id is not None:
            await self._post([self.mount_path, 'role', name, 'secret-id/destroy'], payload={'secret_id': secret_id})
        else:
            await self._post([self.mount_path, 'role', name, 'secret-id-accessor/destroy'], payload={'secret_id_accessor': secret_id_accessor})

    async def login(self, role_id: str, secret_id: Optional[str] = None, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        Authenticate with the AppRole backend


        :param role_id: Role ID
        :param secret_id: Secret ID (if cidr isnt used)
        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        payload = {
            'role_id': role_id
        }
        if secret_id is not None:
            payload['secret_id'] = secret_id

        response = await self._post([self.mount_path, 'login'], payload=payload, wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)


class TokenAuthBackend(HTTPBase):
    """
    Vault Token authentication backend
    """
    async def list_accessors(self, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        List token accessors

        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._list(['auth/token/accessors'], wrap_ttl=wrap_ttl)
        json_data = await response.json()

        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def create(self, name: Optional[str] = None, policies: Optional[List[str]] = None, role: Optional[str] = None,
                     no_parent: bool = False, display_name: Optional[str] = None, meta: Optional[Dict[str, str]] = None,
                     num_uses: int = 0, no_default_policy: bool = False, ttl: Optional[str] = None,
                     orphan: bool = False, wrap_ttl: Optional[int] = None, renewable: Optional[bool] = None,
                     explicit_max_ttl: Optional[int] = None, period: Optional[str] = None) -> ResponseBase:
        """
        Create token

        :param name: Optional token ID
        :param policies: Token policies
        :param role: Role
        :param no_parent: Creates an orphan token
        :param display_name: Token display name
        :param meta: Additional data
        :param num_uses: Number of uses
        :param no_default_policy: If no default policy is to be added
        :param ttl: TTL
        :param orphan: Orphan token
        :param wrap_ttl: Wrap TTL
        :param renewable: Is renewable
        :param explicit_max_ttl: Max TTL
        :param period: TTL period
        :return: Response
        """

        payload = {
            'id': name,
            'policies': policies,
            'meta': meta,
            'no_parent': no_parent,
            'display_name': display_name,
            'num_uses': num_uses,
            'no_default_policy': no_default_policy,
            'renewable': renewable
        }
        if ttl is not None:
            payload['ttl'] = ttl
        elif period is not None:
            payload['perios'] = period
        if explicit_max_ttl is not None:
            payload['explicit_max_ttl'] = explicit_max_ttl

        if orphan:
            response = await self._post('auth/token/create-orphan', payload=payload, wrap_ttl=wrap_ttl)
        elif role is not None:
            response = await self._post(['auth/token/create', role], payload=payload, wrap_ttl=wrap_ttl)
        else:
            response = await self._post('auth/token/create', payload=payload, wrap_ttl=wrap_ttl)

        # TODO possibly set ResponseBase.__getitem__ to use .auth instead of .data for token create responses
        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def lookup(self, token: Optional[str] = None, accessor: Optional[str] = None, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        Lookup token

        :param token: Token ID
        :param accessor: Token accessor
        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        if token is None and accessor is None:
            raise ValueError("Token and Accessor cannot both be none")

        if token is not None:
            # Token Lookup
            response = await self._post('auth/token/lookup', payload={'token': token}, wrap_ttl=wrap_ttl)

            json_data = await response.json()
            return ResponseBase(json_dict=json_data, request_func=self._request)

        # Accessor lookup
        response = await self._post('auth/token/lookup-accessor', payload={'accessor': accessor}, wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def lookup_self(self, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        Lookup active token info

        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._get('auth/token/lookup', wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def renew(self, token: str, increment: Optional[int] = None, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        Renew token

        :param token: Token
        :param increment: Increment to lease
        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        payload = {
            'token': token,
        }
        if increment is not None:
            payload['increment'] = increment

        response = await self._post('auth/token/renew', payload=payload, wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def renew_self(self, increment: Optional[int] = None, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        Renew currently active token

        :param increment: Increment to lease
        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        if increment is not None:
            response = await self._post('auth/token/renew-self', payload={'increment': increment}, wrap_ttl=wrap_ttl)
        else:
            response = await self._post('auth/token/renew-self', payload={}, wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def revoke(self, token: Optional[str] = None, accessor: Optional[str] = None, orphan: bool = False):
        """
        Revoke token

        :param token: Token
        :param accessor: Token accessor
        :param orphan: Orphan token
        """
        if token is None and accessor is None:
            raise ValueError("Token and Accessor cannot both be none")

        if token is not None:
            if orphan:
                await self._post('auth/token/revoke-orphan', payload={'token': token})
            else:
                await self._post('auth/token/revoke', payload={'token': token})
        else:
            await self._post('auth/token/revoke-accessor', payload={'accessor': accessor})

    async def revoke_self(self):
        """
        Revoke currently active token
        """
        await self._post('auth/token/revoke-self', payload=None)

    async def list_roles(self, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        List token roles

        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._list('auth/token/roles', wrap_ttl=wrap_ttl)

        json_data = await response.json()
        return ResponseBase(json_dict=json_data, request_func=self._request)

    async def create_role(self, name: str, allowed_policies: Optional[List[str]] = None, disallowed_policies: Optional[List[str]] = None,
                          orphan: bool = False, period: Optional[str] = None, renewable: bool = False, path_suffix: Optional[str] = None,
                          explicit_max_ttl: Optional[str] = None):
        """
        Create token role

        :param name: Role name
        :param allowed_policies: Allowed policies
        :param disallowed_policies: Disallowed policies, allows you to do "you can have these policies if you dont have these"
        :param orphan: Tokens created with this will be orphans
        :param period: Period TTL
        :param renewable: Is renewable
        :param path_suffix: Role name will have this suffix
        :param explicit_max_ttl: Max TTL
        :return: Response
        """
        payload = {
            'allowed_policies': allowed_policies,
            'disallowed_policies': disallowed_policies,
            'orphan': orphan,
            'period': period,
            'renewable': renewable,
            'path_suffix': path_suffix,
            'explicit_max_ttl': explicit_max_ttl
        }
        await self._post(['auth/token/roles', name], payload=payload)

    async def delete_role(self, name: str):
        """
        Delete token role

        :param name: Role name
        """
        await self._delete(['auth/token/roles', name])

    async def tidy(self):
        """
        Run tidy operation
        """
        await self._post('auth/token/tidy')


class BaseAuth(HTTPBase):
    """
    Base authentication class, basically contains references to the other authentication backends as well as the capability to list the authentication backends.
    """
    def __init__(self, *args, **kwargs) -> None:
        super(BaseAuth, self).__init__(*args, **kwargs)
        self._args = args
        self._kwargs = kwargs

        self._token = None
        self._userpass = None
        self._github = None
        self._approle = None
        self._radius = None
        self._ldap = None

    @property
    def token(self) -> TokenAuthBackend:
        """
        Get Token backend

        :return: Token backend
        """
        if self._token is None:
            self._token = TokenAuthBackend(*self._args, **self._kwargs)

        return self._token

    @property
    def userpass(self) -> UserPassAuthBackend:
        """
        Get UserPass backend

        :return: UserPass backend
        """
        if self._userpass is None:
            self._userpass = UserPassAuthBackend(*self._args, **self._kwargs)

        return self._userpass

    @property
    def github(self) -> GitHubAuthBackend:
        """
        Get GitHub backend

        :return: GitHub backend
        """
        if self._github is None:
            self._github = GitHubAuthBackend(*self._args, **self._kwargs)

        return self._github

    @property
    def approle(self) -> ApproleBackend:
        """
        Get AppRole backend

        :return: AppRole backend
        """
        if self._approle is None:
            self._approle = ApproleBackend(*self._args, **self._kwargs)

        return self._approle

    @property
    def radius(self) -> RadiusAuthBackend:
        """
        Get RADIUS backend

        :return: RADIUS backend
        """
        if self._radius is None:
            self._radius = RadiusAuthBackend(*self._args, **self._kwargs)

        return self._radius

    @property
    def ldap(self) -> LDAPAuthBackend:
        """
        Get LDAP backend

        :return: LDAP backend
        """
        if self._ldap is None:
            self._ldap = LDAPAuthBackend(*self._args, **self._kwargs)

        return self._ldap

    async def list_backends(self, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        List auth backends

        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._get('sys/auth', wrap_ttl=wrap_ttl)

        json = await response.json()
        return ResponseBase(json_dict=json, request_func=self._request)

    def get_userpass_backend(self, mount_path) -> UserPassAuthBackend:
        """
        Get UserPass backend

        :param mount_path: Mount path
        :return: UserPass backend
        """
        return UserPassAuthBackend(*self._args, mount_path=mount_path, **self._kwargs)

    def get_github_backend(self, mount_path) -> GitHubAuthBackend:
        """
        Get GitHub backend

        :param mount_path: Mount path
        :return: GitHub backend
        """
        return GitHubAuthBackend(*self._args, mount_path=mount_path, **self._kwargs)

    def get_approle_backend(self, mount_path) -> ApproleBackend:
        """
        Get AppRole backend

        :param mount_path: Mount path
        :return: AppRole backend
        """
        return ApproleBackend(*self._args, mount_path=mount_path, **self._kwargs)

    def get_radius_backend(self, mount_path) -> RadiusAuthBackend:
        """
        Get RADIUS backend

        :param mount_path: Mount path
        :return: RADIUS backend
        """
        return RadiusAuthBackend(*self._args, mount_path=mount_path, **self._kwargs)

    def get_ldap_backend(self, mount_path) -> LDAPAuthBackend:
        """
        Get LDAP backend

        :param mount_path: Mount path
        :return: LDAP backend
        """
        return LDAPAuthBackend(*self._args, mount_path=mount_path, **self._kwargs)



"""
Base classes
"""
import asyncio
import datetime

from collections import Mapping
from typing import Optional, Union, List, Awaitable, Callable

import aiohttp
import aiohttp.client
import dateutil.parser
import dateutil.relativedelta

from . import exceptions

VAUTL_URL = 'http://localhost:8200'


class HTTPBase(object):
    """
    A generic HTTP class which provides various _get, _list _post etc... methods to query vault
    """
    def __init__(self, vault_url: str = VAUTL_URL, token: Optional[str] = None, verify: bool = True, timeout: int = 10,
                 session: Optional[aiohttp.ClientSession]=None, loop: asyncio.AbstractEventLoop=None):
        self.loop = loop
        if loop is None:
            self.loop = asyncio.get_event_loop()

        self.vault_url = vault_url.rstrip('/')

        self.session = session
        if session is None:
            if not verify:
                connector = aiohttp.TCPConnector(verify_ssl=False, loop=self.loop)
            else:
                connector = None

            self.session = aiohttp.ClientSession(connector=connector, read_timeout=timeout, conn_timeout=timeout, loop=self.loop)  # pylint: disable=unexpected-keyword-arg

        self._auth_token = token
        self.timeout = timeout

    def __del__(self):
        self.close()

    def close(self):
        """
        Close aiohttp session
        """
        if not self.session.closed:
            self.session.close()

    def _make_url(self, path: Union[str, List[str]]) -> str:
        """
        Create a full url to vault

        will join a http prefix like http://vault:5200 with /v1/ and a string or list of paths
        e.g.

        a path of ['auth/user', 'test'] will become http://vault:5200/v1/auth/user/test

        :param path: Vault URL components
        :return: Full vault URL
        """
        if isinstance(path, str):
            return '{0}/v1/{1}'.format(self.vault_url, path.lstrip('/'))
        elif isinstance(path, list):
            return '{0}/v1/{1}'.format(self.vault_url, '/'.join([item.strip('/') for item in path]))
        else:
            raise ValueError("Path isnt a str or list of str")

    async def _get(self, path: Union[str, List[str]], params: Optional[dict] = None,
                   wrap_ttl: Optional[int] = None) -> Awaitable[aiohttp.client.ClientResponse]:
        """
        HTTP GET request

        :param path: Path components
        :param wrap_ttl: Optional TTL
        :return: A response object from aiohttp
        """
        return await self._request('get', path, payload=None, params=params, wrap_ttl=wrap_ttl)

    async def _delete(self, path: Union[str, List[str]], params: Optional[dict] = None,
                      wrap_ttl: Optional[int] = None) -> Awaitable[aiohttp.client.ClientResponse]:
        """
        HTTP DELETE request

        :param path: Path components
        :param wrap_ttl: Optional TTL
        :return: A response object from aiohttp
        """
        return await self._request('delete', path, payload=None, params=params, wrap_ttl=wrap_ttl)

    async def _list(self, path: Union[str, List[str]], params: Optional[dict] = None,
                    wrap_ttl: Optional[int] = None) -> Awaitable[aiohttp.client.ClientResponse]:
        """
        HTTP LIST request

        :param path: Path components
        :param wrap_ttl: Optional TTL
        :return: A response object from aiohttp
        """
        return await self._request('list', path, payload=None, params=params, wrap_ttl=wrap_ttl)

    async def _post(self, path: Union[str, List[str]], payload: Optional[dict] = None, params: Optional[dict] = None,
                    wrap_ttl: Optional[int] = None) -> Awaitable[aiohttp.client.ClientResponse]:
        """
        HTTP POST request

        :param path: Path components
        :param payload: Dictonary of key value to be turned into JSON
        :param wrap_ttl: Optional TTL
        :return: A response object from aiohttp
        """
        return await self._request('post', path, payload=payload, params=params, wrap_ttl=wrap_ttl)

    async def _put(self, path: Union[str, List[str]], payload: Optional[dict] = None, params: Optional[dict] = None,
                   wrap_ttl: Optional[int] = None) -> Awaitable[aiohttp.client.ClientResponse]:
        """
        HTTP PUT request

        :param path: Path components
        :param payload: Dictonary of key value to be turned into JSON
        :param wrap_ttl: Optional TTL
        :return: A response object from aiohttp
        """
        return await self._request('put', path, payload=payload, params=params, wrap_ttl=wrap_ttl)

    async def _request(self, method: str, path: Union[str, List[str]], payload: Optional[dict], params: Optional[dict] = None,
                       wrap_ttl: Optional[int] = None) -> Awaitable[aiohttp.client.ClientResponse]:
        """
        HTTP Request method which takes a method

        It will take care of getting aiohttp to use a HTTP LIST verb

        Adds in the Token and TTL headers, also converts payloads to JSON

        :param method: HTTP Method
        :param path: Path components
        :param payload: Dictonary of key value to be turned into JSON
        :param params: Querystring parameters
        :param wrap_ttl: Optional TTL
        :return: A response object from aiohttp
        """
        url = self._make_url(path)

        headers = {}

        if self._auth_token is not None:
            headers['X-Vault-Token'] = self._auth_token
        if wrap_ttl is not None:
            headers['X-Vault-Wrap-TTL'] = str(wrap_ttl)

        kwargs = {'headers': headers}
        if payload is not None:
            kwargs['json'] = payload

        if params is not None:
            kwargs['params'] = params

        if method != 'list':
            method_func = getattr(self.session, method)(url, **kwargs)
        else:
            # Need to grab some internal stuff from aiohttp to do LIST requests
            method_func = aiohttp.client._RequestContextManager(self.session._request('LIST', url, **kwargs))  # pylint: disable=protected-access

        async with method_func as response:
            return await self._validate_response(response)

    @classmethod
    async def _validate_response(cls, response: aiohttp.client.ClientResponse) -> Awaitable[aiohttp.client.ClientResponse]:
        """
        Takes in a HTTP response, looks through it to see if its legit, if not raise some errors.
        If all is good return the response

        :param response: aiohttp response
        :return: aiohttp response
        """
        if 400 <= response.status < 600:
            if response.headers.get('Content-Type') == 'application/json':
                json_data = await response.json()
                cls._raise_error(response.status, errors=json_data.get('errors'))
            else:
                text = await response.text()
                cls._raise_error(response.status, message=text)
        else:
            return response

    @staticmethod
    def _raise_error(status: int, message: Optional[str] = None, errors: Optional[list] = None):
        """
        Raise an error based on the status code

        :param status: HTTP response status
        :param message: JSON message
        :param errors: JSON error
        :raises exceptions.InvalidRequest: On HTTP 400
        :raises exceptions.Unauthorized: On HTTP 401
        :raises exceptions.Forbidden: On HTTP 403
        :raises exceptions.InvalidPath: On HTTP 404
        :raises exceptions.RateLimitExceeded: On HTTP 429
        :raises exceptions.InternalServerError: On HTTP 500
        :raises exceptions.VaultNotInitialized: On HTTP 501
        :raises exceptions.VaultDown: On HTTP 503
        :raises exceptions.UnexpectedError: On any other HTTP response
        """
        if status == 400:
            raise exceptions.InvalidRequest(message, errors=errors)
        elif status == 401:
            raise exceptions.Unauthorized(message, errors=errors)
        elif status == 403:
            raise exceptions.Forbidden(message, errors=errors)
        elif status == 404:
            raise exceptions.InvalidPath(message, errors=errors)
        elif status == 429:
            raise exceptions.RateLimitExceeded(message, errors=errors)
        elif status == 500:
            raise exceptions.InternalServerError(message, errors=errors)
        elif status == 501:
            raise exceptions.VaultNotInitialized(message, errors=errors)
        elif status == 503:
            raise exceptions.VaultDown(message, errors=errors)
        else:
            raise exceptions.UnexpectedError(message)


class ResponseBase(Mapping):
    """
    A class which wraps the JSON response from Vault

    As it stands the class also implements a read only dictionary type data structure which will
    base its keys of any content inside the 'data' portion of a Vault response. Am undecided yet if
    I want to optionally change that base to 'auth' when thats populated.

    It can also unwrap wrapped responses and rewrap responses, has convenience methods to check expiry etc...
    """
    def __init__(self, json_dict: dict, request_func: Callable[[str, Union[str, List[str]], Optional[dict], Optional[int]],
                                                               Awaitable[aiohttp.client.ClientResponse]]) -> None:
        self._request = request_func

        self.warnings = None
        self.auth = None
        self.renewable = None
        self.lease_duration = None
        self.data = None
        self.wrap_info = None

        self.lease_id = None
        self.request_id = None

        self.wrapped_at = None
        self.expires_at = None

        self._set(json_dict)

    def _set(self, json_dict: dict):
        """
        Sets instance variables to values from a Vault JSON payload

        :param json_dict: JSON dictionary
        """
        self.warnings = json_dict.get('warnings', None)
        self.auth = json_dict.get('auth', None)
        self.renewable = json_dict.get('renewable')
        self.lease_duration = json_dict.get('lease_duration', 0)
        self.data = json_dict.get('data', None)
        self.wrap_info = json_dict.get('wrap_info', None)

        lease_id = json_dict.get('lease_id', '')
        request_id = json_dict.get('request_id', '')
        self.lease_id = lease_id if lease_id != '' else None
        self.request_id = request_id if request_id != '' else None

        if self.data is None:
            self.data = {}

        if self.is_wrapped:
            self.wrapped_at = dateutil.parser.parse(self.wrap_info['creation_time'])
            self.expires_at = self.wrapped_at + dateutil.relativedelta.relativedelta(seconds=self.wrap_info['ttl'])

    @property
    def is_wrapped(self) -> bool:
        """
        Determine if the response is wrapped
        :return: True if the response is wrapped
        """
        return self.wrap_info is not None

    @property
    def wrap_expired(self) -> bool:
        """
        Check if a wrapped response has expired, if the response is not wrapped, it will always return True

        :return: True if the wrapped response has expired
        """
        if self.is_wrapped:
            return datetime.datetime.now(tz=self.wrapped_at.tzinfo) > self.expires_at

        return True

    async def unwrap(self, wrap_ttl: Optional[int] = None):
        """
        Unwrap a wrapped respone else do nothing

        :param wrap_ttl: Wrap TTL
        """
        if self.is_wrapped:
            if self.wrap_expired:
                raise exceptions.VaultWrapExpired()
            payload = {'token': self.wrap_info['token']}
            response = await self._request('post', 'sys/wrapping/unwrap', payload, wrap_ttl=wrap_ttl)
            json_dict = await response.json()

            self._set(json_dict)

    async def rewrap(self):
        """
        Rewrap a wrapped response
        """
        if self.is_wrapped:
            if self.wrap_expired:
                raise exceptions.VaultWrapExpired()
            payload = {'token': self.wrap_info['token']}
            response = await self._request('post', 'sys/wrapping/rewrap', payload)
            json_dict = await response.json()

            self._set(json_dict)

    def __getitem__(self, item):
        return self.data.__getitem__(item)

    def __iter__(self):
        return self.data.__iter__()

    def __len__(self):
        return self.data.__len__()



"""
Collection of exceptions to split out some of the errors Vault could throw
"""


class VaultError(Exception):
    """
    Generic catch-all Vault exception
    """
    def __init__(self, message=None, errors=None):
        if errors:
            message = ', '.join(errors)

        self.errors = errors

        super(VaultError, self).__init__(message)


class InvalidRequest(VaultError):
    """
    Invalid Request Exception
    """


class Unauthorized(VaultError):
    """
    Unauthorized Exception
    """


class Forbidden(VaultError):
    """
    Forbidden Exception
    """


class InvalidPath(VaultError):
    """
    Invalid Path Exception
    """


class RateLimitExceeded(VaultError):
    """
    Rate Limit Exceeded Exception
    """


class InternalServerError(VaultError):
    """
    Internal Server Exception
    """


class VaultNotInitialized(VaultError):
    """
    Vault not initialized exception
    """


class VaultDown(VaultError):
    """
    Vault down exception
    """


class UnexpectedError(VaultError):
    """
    Any other error
    """


class VaultWrapExpired(VaultError):
    """
    Error if trying to unwrap an exception after expiry
    """



"""
/sys classes
"""
from typing import Optional, List
from .base import HTTPBase, ResponseBase


class BaseSys(HTTPBase):
    """
    Collection of methods to act on the /sys/ URLs
    """
    async def seal_status(self) -> ResponseBase:
        """
        Get Vault's seal status

        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._get('sys/seal-status')
        json = await response.json()

        return ResponseBase(json_dict={'data': json}, request_func=self._request)

    async def initialize(self, secret_shares: int = 5, secret_threshold: int = 3, root_token_pgp_key: Optional[str] = None, pgp_keys: Optional[List[str]] = None) -> ResponseBase:
        """
        Initialize Vault

        :param secret_shares: Number of secret keys
        :param secret_threshold: Number of secret keys to unseal
        :param root_token_pgp_key: Root PGP Key
        :param pgp_keys: PGP Keys
        :return: Response
        """
        payload = {
            'secret_shares': secret_shares,
            'secret_threshold': secret_threshold
        }
        if root_token_pgp_key is not None:
            payload['root_token_pgp_key'] = root_token_pgp_key
        if pgp_keys is not None:
            payload['pgp_keys'] = pgp_keys

        response = await self._put('sys/init', payload=payload)
        json = await response.json()

        return ResponseBase(json_dict={'data': json}, request_func=self._request)

    async def remount(self, path_from: str, path_to: str):
        """
        Remount mounted backend to a different path

        :param path_from: Current mount path
        :param path_to: Destination mount path
        """
        payload = {
            'from': path_from,
            'to': path_to
        }
        await self._post('sys/remount', payload=payload)

    async def unseal(self, key: Optional[str]) -> ResponseBase:
        """
        Unseal Vault

        :param key: Unseal key, if str is None then just resets all unlock operations
        """
        if key is not None:
            payload = {'key': key}
        else:
            payload = {'reset': True}

        response = await self._put('sys/unseal', payload=payload)
        json = await response.json()

        return ResponseBase(json_dict={'data': json}, request_func=self._request)

    async def seal(self):
        """
        Seal Vault
        """
        await self._put('sys/seal')

    async def rotate(self):
        """
        Seal Vault
        """
        await self._put('sys/rotate')

    async def health(self) -> ResponseBase:
        """
        Vault health
        """
        params = {
            'standbyok': 'true',
            'activecode': 200,
            'standbycode': 200,
            'sealedcode': 200,
            'uninitcode': 200
        }

        response = await self._get('sys/health', params=params)
        json = await response.json()

        return ResponseBase(json_dict={'data': json}, request_func=self._request)



"""
Main client module
"""
from typing import Optional

from . import exceptions
from .base import HTTPBase
from .secrets import BaseSecret
from .audit import BaseAudit
from .vault_sys import BaseSys
from .policy import BasePolicy
from .auth import BaseAuth


class VaultClient(HTTPBase):
    """
    Vault client, keeps a session and facilitates access to vault backends
    """
    def __init__(self, *args, **kwargs):
        super(VaultClient, self).__init__(*args, **kwargs)

        # If session is None then one would have been generated, pass that one round
        kwargs.pop('session', None)
        self.secrets = BaseSecret(*args, session=self.session, **kwargs)
        self.audit = BaseAudit(*args, session=self.session, **kwargs)
        self.sys = BaseSys(*args, session=self.session, **kwargs)
        self.policies = BasePolicy(*args, session=self.session, **kwargs)
        self.auth = BaseAuth(*args, session=self.session, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    async def is_authenticated(self) -> bool:
        """
        Attempt to call `auth/token/lookup-self`, if it's successful then return True and otherwise return False

        :return: True if token is valid, False otherwise
        """
        if self._auth_token is None:
            return False

        try:
            await self._get('auth/token/lookup-self')
            # If it doesnt raise an exception
            return True
        except exceptions.Forbidden:
            return False
        except exceptions.InvalidPath:
            return False
        except exceptions.InvalidRequest:
            return False

    def logout(self):
        """
        Removes the current token from the client
        """
        self._auth_token = None

    @property
    def auth_token(self) -> Optional[str]:
        """
        Get current auth token

        :return: Auth token
        """
        return self._auth_token

    @auth_token.setter
    def auth_token(self, token: str):
        """
        Set auth token

        :param token: Auth token
        """
        self._auth_token = token



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `aiovault` package."""
import os
import asyncio
import pytest
import aiovault
import aiovault.base
import binascii
import base64

from .common import BaseTestCase, ConsulTestCase


class TestBaseSecretBackend(BaseTestCase):
    async def test_list_secret_backends(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        result = await client.secrets.list()

        assert 'secret/' in result
        assert 'cubbyhole/' in result
        assert 'sys/' in result

    async def test_mount_generic_backend(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.generic.mount(
            path='test_generic',
            description='test_mount'
        )

        result = await client.secrets.list()

        assert 'test_generic/' in result
        assert 'test_mount' in result['test_generic/']['description']

    async def test_get_mounted_generic_backend(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.generic.mount(path='test_generic2')

        new_mount = client.secrets.get_generic_secret_backend('test_generic2')

        await new_mount.create('test_mount', key1='value1')
        secret = await new_mount.read('test_mount')

        assert isinstance(secret, aiovault.base.ResponseBase)
        assert 'key1' in secret

        # Check its not in the generic /secret backend
        with pytest.raises(aiovault.exceptions.InvalidPath):
            await client.secrets.generic.read('test_mount')


class TestGenericSecretBackend(BaseTestCase):
    async def test_create_secret(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.generic.create('test/create/test1', value='test')
        await client.secrets.generic.create('test/create/test2', value='test')

    async def test_read_secret(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.generic.create('test/read/test1', key1='value1', key2='value2')

        secret = await client.secrets.generic.read('test/read/test1')

        assert isinstance(secret, aiovault.base.ResponseBase)
        assert 'key1' in secret
        assert 'key2' in secret
        assert secret['key1'] == 'value1'
        assert secret['key2'] == 'value2'

    async def test_list_secret(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.generic.create('test/list/test1', key1='value1')
        await client.secrets.generic.create('test/list/test2', key2='value2')
        await client.secrets.generic.create('test/list/test3/test4', key2='value2')

        secret = await client.secrets.generic.list('test/list')

        # Should show us keys test1, test2 and directory test3/
        assert isinstance(secret, aiovault.base.ResponseBase)
        assert 'keys' in secret
        for key_name in ('test1', 'test2', 'test3/'):
            assert key_name in secret['keys']

    async def test_secret_ttl(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)
        # Generic secret backend will not automatically remove values

        await client.secrets.generic.create('test/ttl/test1', key1='value1', ttl=3)

        secret = await client.secrets.generic.read('test/ttl/test1')

        assert 'ttl' in secret
        assert secret['ttl'] == 3
        assert secret.lease_duration == 3

    async def test_update_secret(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.generic.create('test/update/test1', key1='value1')
        # Normal vault will do an overwrite on update with generic secret backend
        await client.secrets.generic.update('test/update/test1', key2='value2')

        secret = await client.secrets.generic.read('test/update/test1')

        assert 'key1' in secret
        assert 'key2' in secret
        assert secret['key1'] == 'value1'
        assert secret['key2'] == 'value2'

    async def test_delete_secret(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.generic.create('test/read/test1', key1='value1')

        secret = await client.secrets.generic.read('test/read/test1')

        assert 'key1' in secret

        await client.secrets.generic.delete('test/read/test1')

        with pytest.raises(aiovault.exceptions.InvalidPath):
            await client.secrets.generic.read('test/read/test1')

    async def test_list_wrapped_secret(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.generic.create('test/list/test1', key1='value1')
        await client.secrets.generic.create('test/list/test2', key2='value2')
        await client.secrets.generic.create('test/list/test3/test4', key2='value2')

        secret = await client.secrets.generic.list('test/list', wrap_ttl=60)

        assert secret.is_wrapped

        await secret.unwrap()

        assert not secret.is_wrapped
        assert 'keys' in secret
        for key_name in ('test1', 'test2', 'test3/'):
            assert key_name in secret['keys']

    async def test_read_wrapped_secret(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.generic.create('test/read/test1', key1='value1', key2='value2')

        secret = await client.secrets.generic.read('test/read/test1', wrap_ttl=60)

        assert secret.is_wrapped

        await secret.unwrap()

        assert not secret.is_wrapped
        assert 'key1' in secret
        assert 'key2' in secret
        assert secret['key1'] == 'value1'
        assert secret['key2'] == 'value2'

    @pytest.mark.skipif('SLOW_TESTS' not in os.environ, reason="Skipping slow tests by default")
    async def test_wrapped_secret_expire(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.generic.create('test/read/test1', key1='value1', key2='value2')

        secret = await client.secrets.generic.read('test/read/test1', wrap_ttl=1)

        assert secret.is_wrapped

        await asyncio.sleep(2, loop=loop)

        with pytest.raises(aiovault.exceptions.VaultWrapExpired):
            await secret.unwrap()

    async def test_read_rewrapped_secret(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.generic.create('test/read_wrapped/test1', key1='value1')

        secret = await client.secrets.generic.read('test/read_wrapped/test1', wrap_ttl=60)

        assert secret.is_wrapped
        initial_token = secret.wrap_info['token']

        await secret.rewrap()
        assert initial_token != secret.wrap_info['token']

        await secret.unwrap()

        assert not secret.is_wrapped
        assert 'key1' in secret
        assert secret['key1'] == 'value1'


class TestTransitSecretBackend(BaseTestCase):
    @pytest.mark.run(order=1)
    async def test_mount_transit(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.transit.mount('transit', description='test_mount_transit')
        result = await client.secrets.list()
        assert 'transit/' in result

    async def test_create_key(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.transit.create('test_create_key', key_type='ecdsa-p256')

        result = await client.secrets.transit.read('test_create_key')

        assert 'name' in result
        assert result['name'] == 'test_create_key'
        assert 'type' in result
        assert result['type'] == 'ecdsa-p256'

    async def test_list_key(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.transit.create('test_list_key')

        result = await client.secrets.transit.list()

        assert 'keys' in result
        assert 'test_list_key' in result['keys']

    async def test_update_key(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.transit.create('test_update_key', key_type='ecdsa-p256')
        await client.secrets.transit.update('test_update_key', deletion_allowed=True)

        result = await client.secrets.transit.read('test_update_key')

        assert result['name'] == 'test_update_key'
        assert result['deletion_allowed']

    async def test_delete_key(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.transit.create('test_delete_key')

        result = await client.secrets.transit.list()
        assert 'test_delete_key' in result['keys']

        with pytest.raises(aiovault.exceptions.InvalidRequest):
            await client.secrets.transit.delete('test_delete_key')

        await client.secrets.transit.update('test_delete_key', deletion_allowed=True)
        await client.secrets.transit.delete('test_delete_key')

        result = await client.secrets.transit.list()
        assert 'test_delete_key' not in result['keys']

    async def test_rotate_key(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.transit.create('test_rotate_key')

        before_rotate = await client.secrets.transit.read('test_rotate_key')
        await client.secrets.transit.rotate('test_rotate_key')
        after_rotate = await client.secrets.transit.read('test_rotate_key')

        assert '1' in before_rotate['keys']
        assert '2' not in before_rotate['keys']
        assert '1' in after_rotate['keys']
        assert '2' in after_rotate['keys']

    async def test_export_key(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.transit.create('test_export_key', exportable=True)

        result = await client.secrets.transit.export('test_export_key', 'encryption-key')

        assert 'name' in result
        assert 'keys' in result
        assert 'type' in result

    async def test_encryption(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.transit.create('test_enc_key', exportable=True)

        result = await client.secrets.transit.encrypt('test_enc_key', 'test1234')

        assert 'ciphertext' in result

    async def test_batch_encryption(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.transit.create('test_batch_enc_key', exportable=True)

        encode = [{'plaintext': 'test1234'}, {'plaintext': '4321dcba'}]
        result = await client.secrets.transit.encrypt('test_batch_enc_key', '', batch_input=encode)

        assert 'batch_results' in result

    async def test_decryption(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.transit.create('test_dec_key', exportable=True)

        enc_result = await client.secrets.transit.encrypt('test_dec_key', 'test1234')

        dec_result = await client.secrets.transit.decrypt('test_dec_key', enc_result['ciphertext'])

        assert 'plaintext' in dec_result
        assert dec_result['plaintext'] == 'test1234'

    async def test_batch_decryption(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.transit.create('test_batch_dec_key', exportable=True)

        encode = [{'plaintext': 'test1234'}, {'plaintext': '4321dcba'}]
        enc_result = await client.secrets.transit.encrypt('test_batch_dec_key', '', batch_input=encode)

        dec_result = await client.secrets.transit.decrypt('test_batch_dec_key', '', batch_input=enc_result['batch_results'])

        for pre_enc, post_dec in zip(encode, dec_result['batch_results']):
            assert pre_enc['plaintext'] == post_dec['plaintext']

    async def test_rewrap(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.transit.create('test_rewrap_key', exportable=True)

        enc_result = await client.secrets.transit.encrypt('test_rewrap_key', 'test1234')
        assert enc_result['ciphertext'].startswith('vault:v1')

        await client.secrets.transit.rotate('test_rewrap_key')
        wrapped_result = await client.secrets.transit.rewrap('test_rewrap_key', enc_result['ciphertext'])
        assert wrapped_result['ciphertext'].startswith('vault:v2')

        dec_result = await client.secrets.transit.decrypt('test_rewrap_key', wrapped_result['ciphertext'])
        assert dec_result['plaintext'] == 'test1234'

    async def test_rewrap_batch(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.transit.create('test_batch_rewrap_key', exportable=True)

        encode = [{'plaintext': 'test1234'}, {'plaintext': '4321dcba'}]
        enc_result = await client.secrets.transit.encrypt('test_batch_rewrap_key', '', batch_input=encode)
        await client.secrets.transit.rotate('test_batch_rewrap_key')
        wrapped_result = await client.secrets.transit.rewrap('test_batch_rewrap_key', '', batch_input=enc_result['batch_results'])

        dec_result = await client.secrets.transit.decrypt('test_batch_rewrap_key', '', batch_input=wrapped_result['batch_results'])
        for pre_enc, post_dec in zip(encode, dec_result['batch_results']):
            assert pre_enc['plaintext'] == post_dec['plaintext']

    async def test_generate_key_data(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.transit.create('test_get_data_key')

        wrapped_data = await client.secrets.transit.generate_key_data('test_get_data_key')
        assert 'ciphertext' in wrapped_data

        plaintext_data = await client.secrets.transit.generate_key_data('test_get_data_key', result_type='plaintext')
        assert 'ciphertext' in plaintext_data
        assert 'plaintext' in plaintext_data

    async def test_generate_random_data(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        bytes_result = await client.secrets.transit.generate_random_bytes(length=32, result_format='bytes')
        hex_result = await client.secrets.transit.generate_random_bytes(length=8, result_format='hex')
        base64_result = await client.secrets.transit.generate_random_bytes(length=16, result_format='base64')

        assert isinstance(bytes_result['random_bytes'], bytes)
        hex_data = binascii.unhexlify(hex_result['random_bytes'])
        assert len(hex_data) == 8
        base64_data = base64.b64decode(base64_result['random_bytes'].encode())
        assert len(base64_data) == 16

    async def test_generate_hmac(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.transit.create('test_hmac_key')

        result = await client.secrets.transit.generate_hmac('test_hmac_key', 'test1234', result_format='base64')
        assert 'hmac' in result

    async def test_sign_data(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.transit.create('test_sign_key', key_type='ecdsa-p256')

        result = await client.secrets.transit.sign('test_sign_key', 'test1234')
        assert 'signature' in result

    async def test_verify_signature(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.transit.create('test_verify_sig_key', key_type='ecdsa-p256')

        sig_result = await client.secrets.transit.sign('test_verify_sig_key', 'test1234', algorithm='sha2-256')
        sig_verify_result = await client.secrets.transit.verify('test_verify_sig_key', 'test1234', signature=sig_result['signature'], algorithm='sha2-256')

        assert sig_verify_result['valid']

    async def test_verify_hmac(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.transit.create('test_verify_hmac_key', key_type='ecdsa-p256')

        hmac_result = await client.secrets.transit.generate_hmac('test_verify_hmac_key', 'test1234', algorithm='sha2-256')
        hmac_verify_result = await client.secrets.transit.verify('test_verify_hmac_key', 'test1234', hmac=hmac_result['hmac'], algorithm='sha2-256')

        assert hmac_verify_result['valid']


class TestTOTPSecretBackend(BaseTestCase):
    @pytest.mark.run(order=1)
    async def test_mount_totp(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.totp.mount('totp', description='test_mount_totp')
        result = await client.secrets.list()
        assert 'totp/' in result

    async def test_create_secret(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.totp.create('test_create', generate=True, issuer='VaultAPI', account_name='Test')

        result = await client.secrets.totp.list()
        assert 'keys' in result
        assert 'test_create' in result['keys']

    async def test_read_secret(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.totp.create('test_read', generate=True, issuer='VaultAPI', account_name='Test')

        result = await client.secrets.totp.read('test_read')
        assert 'algorithm' in result
        assert 'digits' in result
        assert result['algorithm'] == 'SHA1'
        assert result['digits'] == 6

    async def test_delete_secret(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.totp.create('test_delete', generate=True, issuer='VaultAPI', account_name='Test')

        result = await client.secrets.totp.list()
        assert 'test_delete' in result['keys']

        await client.secrets.totp.delete('test_delete')

        result = await client.secrets.totp.list()
        assert 'test_delete' not in result['keys']

    async def test_generate_code(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.totp.create('test_generate', generate=True, issuer='VaultAPI', account_name='Test')

        result = await client.secrets.totp.generate_code('test_generate')
        assert 'code' in result

    async def test_validate_code(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.totp.create('test_verify', generate=True, issuer='VaultAPI', account_name='Test')

        code_result = await client.secrets.totp.generate_code('test_verify')

        verify_result = await client.secrets.totp.validate_code('test_verify', code_result['code'])
        assert 'valid' in verify_result
        assert verify_result['valid']


class TestConsulSecretBackend(ConsulTestCase):
    consul_policy = 'key "" { policy = "read" }'

    @pytest.mark.run(order=1)
    async def test_mount_consul(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.consul.mount('consul', address='127.0.0.1', port=8500, management_token='master_token', description='test_mount_totp', https=False)
        result = await client.secrets.list()
        assert 'consul/' in result

    async def test_create_role(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.consul.create('test_create', consul_policy=self.consul_policy)

        result = await client.secrets.consul.list()
        assert 'keys' in result
        assert 'test_create' in result['keys']

    async def test_read_role(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.consul.create('test_read', consul_policy=self.consul_policy)

        result = await client.secrets.consul.read('test_read')
        assert 'policy' in result
        assert 'lease' in result
        assert 'token_type' in result
        assert result['token_type'] == 'client'

    async def test_delete_role(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.consul.create('test_delete', consul_policy=self.consul_policy)

        result = await client.secrets.consul.list()
        assert 'test_delete' in result['keys']

        await client.secrets.consul.delete('test_delete')

        result = await client.secrets.consul.list()
        assert 'test_delete' not in result['keys']

    async def test_generate_credential(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.consul.create('test_generate', consul_policy=self.consul_policy)

        result = await client.secrets.consul.generate_credential('test_generate')
        assert 'token' in result



from tests import run_vault


class BaseTestCase(object):
    proc = None
    config = None
    consul_config = None

    @classmethod
    def setup_class(cls):
        cls.proc = run_vault.VaultProc(config=cls.config, consul_config=cls.consul_config)
        cls.proc.run()

        return cls()

    @classmethod
    def teardown_class(cls):
        cls.proc.stop()


class NoTLSTestCase(BaseTestCase):
    config = 'vault_no_tls.hcl'


class ConsulTestCase(BaseTestCase):
    consul_config = 'consul_server.json'



#!/usr/bin/env python3
import os
import subprocess
import shutil
import sys
import time
import zipfile

import requests

VAULT_CONF_FOLDER = os.path.join(os.path.dirname(__file__), 'config')
VAULT_VERSION = "0.7.3"
CONSUL_VERSION = "0.9.0"


class VaultProc:
    VAULT_ZIP = "vault_{0}_linux_amd64.zip".format(VAULT_VERSION)
    VAULT_DOWNLOAD_URL = "https://releases.hashicorp.com/vault/{0}/{1}".format(VAULT_VERSION, VAULT_ZIP)
    CONSUL_ZIP = "consul_{0}_linux_amd64.zip".format(CONSUL_VERSION)
    CONSUL_DOWNLOAD_URL = "https://releases.hashicorp.com/consul/{0}/{1}".format(CONSUL_VERSION, CONSUL_ZIP)

    def __init__(self, config=None, consul_config=None):
        self._get_vault()

        if config is None:
            self.config = None
        else:
            self.config = os.path.join(VAULT_CONF_FOLDER, config)

        if consul_config is None:
            self.consul_config = None
        else:
            self.consul_config = os.path.join(VAULT_CONF_FOLDER, consul_config)
            self._get_consul()

        self.unseal_keys = None
        self.root_token = None
        self.vault_proc = None
        self.consul_proc = None
        self.consul_master_acl_token = 'master_token'

    def _get_vault(self):
        if not os.path.exists('./vault'):
            response = requests.get(self.VAULT_DOWNLOAD_URL, stream=True)
            if response.status_code != 200:
                print("Failed to download vault")
                sys.exit(1)

            with open('vault.zip', 'wb') as vault_zip:
                response.raw.decode_content = True
                shutil.copyfileobj(response.raw, vault_zip)

            zip_obj = zipfile.ZipFile('vault.zip')
            zip_obj.extract('vault')

            os.remove('vault.zip')
            os.chmod('./vault', 0o0755)

    def _get_consul(self):
        if not os.path.exists('./consul'):
            response = requests.get(self.CONSUL_DOWNLOAD_URL, stream=True)
            if response.status_code != 200:
                print("Failed to download consul")
                sys.exit(1)

            with open('consul.zip', 'wb') as consul_zip:
                response.raw.decode_content = True
                shutil.copyfileobj(response.raw, consul_zip)

            zip_obj = zipfile.ZipFile('consul.zip')
            zip_obj.extract('consul')

            os.remove('consul.zip')
            os.chmod('./consul', 0o0755)

    def run(self, quiet=True):
        if self.config is not None:
            self._run_prod(quiet)
        else:
            self._run_dev(quiet)

    def _run_prod(self, quiet=True):
        if self.consul_config is not None:
            self.consul_proc = subprocess.Popen(['./consul', 'agent', '-dev', '-config-file={0}'.format(self.consul_config)], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        self.vault_proc = subprocess.Popen(['./vault', 'server', '-config={0}'.format(self.config)], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        if not quiet:
            print("Vault started")
            print("No keys as not initialised")

    def _run_dev(self, quiet=True):
        if self.consul_config is not None:
            self.consul_proc = subprocess.Popen(['./consul', 'agent', '-dev', '-config-file={0}'.format(self.consul_config)], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        self.vault_proc = subprocess.Popen(['./vault', 'server', '-dev'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        while True:
            line = self.vault_proc.stdout.readline().decode()
            if line.startswith('Unseal Key'):
                self.unseal_key = line.strip().split(': ')[1]
                continue

            if line.startswith('Root Token'):
                self.root_token = line.strip().split(': ')[1]
                break

        if not quiet:
            print("Vault started")
            print("Unseal key: {0}\nRoot token: {1}".format(self.unseal_key, self.root_token))

    def run_interactive(self):
        self.run(quiet=False)
        print("Waiting for Ctrl+C")

        try:
            while True:
                time.sleep(10)
        except KeyboardInterrupt:
            pass

        self.stop()

    def stop(self):
        if self.vault_proc is not None and self.vault_proc.poll() is None:
            self.vault_proc.kill()

        if self.consul_proc is not None and self.consul_proc.poll() is None:
            self.consul_proc.kill()

    def __del__(self):
        self.stop()


if __name__ == '__main__':
    vault_proc = VaultProc()
    vault_proc.run_interactive()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `aiovault` package."""
import os
import pytest
import aiovault
import aiovault.base

from .common import BaseTestCase


class TestBaseAuth(BaseTestCase):
    async def test_list_auth_backends(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        result = await client.auth.list_backends()
        assert 'token/' in result


class TestGitHubAuth(BaseTestCase):
    @pytest.mark.run(order=1)
    async def test_mount_userpass(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.github.mount('github', organization='openrazer', description='test_mount')
        result = await client.auth.list_backends()
        assert 'github/' in result
        assert result['github/']['description'] == 'test_mount'

    async def test_map_team_policy(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.github.map_team_policy('test_map', ['default'])
        result = await client.auth.github.list_teams()
        assert 'keys' in result
        assert 'test_map' in result['keys']

    async def test_get_team(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.github.map_team_policy('test_read', ['default'])
        result = await client.auth.github.get_team('test_read')
        assert 'value' in result
        assert result['value'] == 'default'

    async def test_del_team(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.github.map_team_policy('test_del', ['default'])
        result = await client.auth.github.get_team('test_del')
        assert 'value' in result

        await client.auth.github.del_team('test_del')

        result = await client.auth.github.get_team('test_del')
        assert 'value' not in result

    async def test_map_user_policy(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.github.map_user_policy('test_map_user', ['default'])
        result = await client.auth.github.list_teams()
        assert 'keys' in result
        assert 'test_map' in result['keys']

    async def test_get_user(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.github.map_user_policy('test_read_user', ['default'])
        result = await client.auth.github.get_user('test_read_user')
        assert 'value' in result
        assert result['value'] == 'default'

    async def test_del_user(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.github.map_user_policy('test_del_user', ['default'])
        result = await client.auth.github.get_user('test_del_user')
        assert 'value' in result

        await client.auth.github.del_user('test_del_user')

        result = await client.auth.github.get_user('test_del_user')
        assert 'value' not in result

    @pytest.mark.skipif('TEST_GITHUB_OAUTH_KEY' not in os.environ, reason="Skipping as no GitHub key")
    async def test_login(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.github.map_user_policy('terrycain', ['default'])
        token_obj = await client.auth.github.login(os.environ['TEST_GITHUB_OAUTH_KEY'])
        assert 'client_token' in token_obj.auth
        assert 'metadata' in token_obj.auth
        assert token_obj.auth['metadata']['username'] == 'terrycain'


class TestRadiusAuth(BaseTestCase):
    @pytest.mark.run(order=1)
    async def test_mount_radius(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        radius_host = os.environ.get('TEST_RADIUS_HOST', 'localhost')  # Because my radius server isnt on localhost
        await client.auth.radius.mount('radius', radius_host=radius_host, secret='testing123', unregistered_user_policies=['default'], description='test_mount')
        result = await client.auth.list_backends()
        assert 'radius/' in result
        assert result['radius/']['description'] == 'test_mount'

    async def test_get_backend(self, loop):  # For completeness's sake
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)
        radius_backend = client.auth.get_radius_backend('radius')

        await radius_backend.create('test_get_backend_user')

        result = await radius_backend.list()
        assert 'keys' in result
        assert 'test_get_backend_user' in result['keys']

    async def test_create_user(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.radius.create('test_create_user')

        result = await client.auth.radius.list()
        assert 'keys' in result
        assert 'test_create_user' in result['keys']

    async def test_read_user(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.radius.create('test_read_user')

        result = await client.auth.radius.read('test_read_user')
        assert 'policies' in result

    async def test_delete_user(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.radius.create('test_delete_user')

        result = await client.auth.radius.list()
        assert 'test_delete_user' in result['keys']

        await client.auth.radius.delete('test_delete_user')

        result = await client.auth.radius.list()
        assert 'test_delete_user' not in result['keys']

    async def test_login(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        result = await client.auth.radius.login('vaulttest', 'test1234')
        assert 'client_token' in result.auth
        assert 'accessor' in result.auth


class TestUserPassAuth(BaseTestCase):
    @pytest.mark.run(order=1)
    async def test_mount_userpass(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.userpass.mount('userpass')
        result = await client.auth.list_backends()
        assert 'userpass/' in result

    async def test_create_user(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.userpass.create('test_create', 'test1234')
        result = await client.auth.userpass.list()
        assert 'keys' in result
        assert 'test_create' in result['keys']

    async def test_read_user(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.userpass.create('test_read', 'test1234')
        result = await client.auth.userpass.read('test_read')
        assert 'policies' in result
        assert result['policies'] == ''

    async def test_update_user(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.userpass.create('test_update', 'test1234')
        await client.auth.userpass.update('test_update', policies=['default'])
        result = await client.auth.userpass.read('test_update')
        assert 'policies' in result
        assert result['policies'] == 'default'

    async def test_update_policies(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.userpass.create('test_update_pol', 'test1234')
        await client.auth.userpass.update_policies('test_update_pol', policies=['default'])
        result = await client.auth.userpass.read('test_update_pol')
        assert 'policies' in result
        assert result['policies'] == 'default'

    async def test_delete_user(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.userpass.create('test_delete', 'test1234')
        await client.auth.userpass.delete('test_delete')
        with pytest.raises(aiovault.exceptions.InvalidPath):
            await client.auth.userpass.read('test_delete')

    async def test_login(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.userpass.create('test_login', 'test1234')
        token_obj = await client.auth.userpass.login('test_login', 'test1234')

        # TODO perhaps point __getitem__ to auth
        assert 'client_token' in token_obj.auth
        assert 'metadata' in token_obj.auth
        assert token_obj.auth['metadata']['username'] == 'test_login'

    async def test_update_password(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.userpass.create('test_updatepw', 'test1234')
        await client.auth.userpass.update_password('test_updatepw', 'password')

        # TODO perhaps point __getitem__ to auth
        token_obj = await client.auth.userpass.login('test_updatepw', 'password')
        assert 'client_token' in token_obj.auth


class TestTokenAuth(BaseTestCase):
    async def test_list_accessor(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        result = await client.auth.token.list_accessors()
        assert 'keys' in result
        assert isinstance(result['keys'], list)

    async def test_create_token(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        result = await client.auth.token.create(display_name='test_create')
        assert result.auth is not None
        assert 'client_token' in result.auth
        assert 'accessor' in result.auth

    async def test_create_orphan_token(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        result = await client.auth.token.create(orphan=True, display_name='test_orphan')
        assert result.auth is not None
        assert 'client_token' in result.auth
        assert 'accessor' in result.auth

    async def test_token_lookup(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        token = await client.auth.token.create(display_name='test-lookup')
        token_id = token.auth['client_token']

        lookup_result = await client.auth.token.lookup(token=token_id)
        assert token_id == lookup_result['id']
        assert lookup_result['display_name'] == 'token-' + 'test-lookup'

    async def test_token_lookup_accessor(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        token = await client.auth.token.create(display_name='test-lookup-accessor')
        accessor_id = token.auth['accessor']

        lookup_result = await client.auth.token.lookup(accessor=accessor_id)
        assert accessor_id == lookup_result['accessor']
        assert lookup_result['display_name'] == 'token-' + 'test-lookup-accessor'

    async def test_lookup_self(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        lookup_result = await client.auth.token.lookup_self()
        assert lookup_result['display_name'] == 'root'
        assert lookup_result['id'] == self.proc.root_token

    async def test_renew(self, loop):
        root_client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        token = await root_client.auth.token.create(ttl=60, renewable=True)
        lookup_result = await root_client.auth.token.lookup(token=token.auth['client_token'])
        await root_client.auth.token.renew(token.auth['client_token'], increment='120s')
        lookup_result2 = await root_client.auth.token.lookup(token=token.auth['client_token'])

        assert lookup_result['expire_time'] != lookup_result2['expire_time']

    async def test_renew_self(self, loop):
        with aiovault.VaultClient(token=self.proc.root_token, loop=loop) as root_client:
            token = await root_client.auth.token.create(ttl=60, renewable=True)

        new_client = aiovault.VaultClient(token=token.auth['client_token'], loop=loop)

        lookup_result = await new_client.auth.token.lookup_self()
        await new_client.auth.token.renew_self()
        lookup_result2 = await new_client.auth.token.lookup_self()

        assert lookup_result['expire_time'] != lookup_result2['expire_time']

    async def test_revoke(self, loop):
        root_client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        token = await root_client.auth.token.create(ttl=60, renewable=True)
        await root_client.auth.token.revoke(token=token.auth['client_token'])

        with pytest.raises(aiovault.exceptions.Forbidden):
            await root_client.auth.token.lookup(token=token.auth['client_token'])

    async def test_revoke_accessor(self, loop):
        root_client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        token = await root_client.auth.token.create(ttl=60, renewable=True)
        await root_client.auth.token.revoke(accessor=token.auth['accessor'])

        with pytest.raises(aiovault.exceptions.Forbidden):
            await root_client.auth.token.lookup(token=token.auth['client_token'])

    async def test_revoke_self(self, loop):
        root_client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        token = await root_client.auth.token.create(ttl=60, renewable=True)
        client = aiovault.VaultClient(token=token.auth['client_token'], loop=loop)
        await client.auth.token.revoke_self()

        with pytest.raises(aiovault.exceptions.Forbidden):
            await root_client.auth.token.lookup(token=token.auth['client_token'])

    async def test_revoke_and_orphan(self, loop):
        root_client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        token = await root_client.auth.token.create(ttl=120, renewable=True)
        client = aiovault.VaultClient(token=token.auth['client_token'], loop=loop)
        sub_token = await client.auth.token.create(ttl=120, renewable=True)

        await root_client.auth.token.revoke(token=token.auth['client_token'], orphan=True)

        await root_client.auth.token.lookup(token=sub_token.auth['client_token'])

        with pytest.raises(aiovault.exceptions.Forbidden):
            await root_client.auth.token.lookup(token=token.auth['client_token'])

    async def test_create_role(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.token.create_role('test_create_role')

        # list_roles raises 404 if no roles
        result = await client.auth.token.list_roles()
        assert 'keys' in result
        assert 'test_create_role' in result['keys']

    async def test_delete_role(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.token.create_role('test_create_role')
        await client.auth.token.list_roles()
        await client.auth.token.delete_role('test_create_role')
        with pytest.raises(aiovault.exceptions.InvalidPath):
            await client.auth.token.list_roles()

    async def test_tidy(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.token.tidy()


class TestAppRoleAuth(BaseTestCase):
    @pytest.mark.run(order=1)
    async def test_mount_approle(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.approle.mount('approle', description='test_mount')
        result = await client.auth.list_backends()
        assert 'approle/' in result
        assert result['approle/']['description'] == 'test_mount'

    async def test_approle_list(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.approle.create('test_create')

        result = await client.auth.approle.list()
        assert 'keys' in result
        assert 'test_create' in result['keys']

    async def test_approle_read(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.approle.create('test_read')

        result = await client.auth.approle.read('test_read')
        assert 'bind_secret_id' in result
        assert 'policies' in result
        assert 'default' in result['policies']

    async def test_approle_update(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.approle.create('test_update')
        await client.auth.approle.update('test_update', bind_secret_id=False, bound_cidr_list=['172.16.30.0/24', '172.16.31.0/24'])
        result = await client.auth.approle.read('test_update')
        assert not result['bind_secret_id']
        assert result['bound_cidr_list'] == '172.16.30.0/24,172.16.31.0/24'

    async def test_approle_delete(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.approle.create('test_delete')
        await client.auth.approle.read('test_delete')
        await client.auth.approle.delete('test_delete')

        with pytest.raises(aiovault.exceptions.InvalidPath):
            await client.auth.approle.read('test_delete')

    async def test_approle_read_role_id(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.approle.create('test_read_role_id')
        result = await client.auth.approle.read_roleid('test_read_role_id')
        assert 'role_id' in result

    async def test_approle_set_role_id(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.approle.create('test_set_role_id')
        await client.auth.approle.set_roleid('test_set_role_id', 'new_role_id')
        result = await client.auth.approle.read_roleid('test_set_role_id')
        assert 'role_id' in result
        assert result['role_id'] == 'new_role_id'

    async def test_approle_generate_secret_id(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.approle.create('test_gen_secret_id')
        result = await client.auth.approle.generate_secret_id('test_gen_secret_id')
        assert 'secret_id_accessor' in result
        assert 'secret_id' in result

    async def test_approle_generate_custom_secret_id(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.approle.create('test_custom_secret_id')
        result = await client.auth.approle.generate_secret_id('test_custom_secret_id', custom_id='test1234')
        assert 'secret_id_accessor' in result
        assert 'secret_id' in result
        assert result['secret_id'] == 'test1234'

    async def test_approle_list_secret_id(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.approle.create('test_list_secret_id')
        sec_id_result = await client.auth.approle.generate_secret_id('test_list_secret_id')

        result = await client.auth.approle.list_secret_ids('test_list_secret_id')
        assert 'keys' in result
        assert sec_id_result['secret_id_accessor'] in result['keys']

    async def test_approle_lookup_secret_id(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.approle.create('test_lookup_secret_id')
        sec_id_result = await client.auth.approle.generate_secret_id('test_lookup_secret_id', metadata={'test1': 'value1'})

        result = await client.auth.approle.lookup_secret_id('test_lookup_secret_id', secret_id=sec_id_result['secret_id'])
        assert 'creation_time' in result
        assert 'metadata' in result
        assert result['metadata']['test1'] == 'value1'

    async def test_approle_lookup_secret_id_accessor(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.approle.create('test_lookup_secret_id_accessor')
        sec_id_result = await client.auth.approle.generate_secret_id('test_lookup_secret_id_accessor', metadata={'test2': 'value2'})

        result = await client.auth.approle.lookup_secret_id('test_lookup_secret_id_accessor', secret_id_accessor=sec_id_result['secret_id_accessor'])
        assert 'creation_time' in result
        assert 'metadata' in result
        assert result['metadata']['test2'] == 'value2'

    async def test_approle_destroy_secret_id(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.approle.create('test_destroy_secret_id')
        sec_id_result = await client.auth.approle.generate_secret_id('test_destroy_secret_id', metadata={'test3': 'value3'})

        await client.auth.approle.destroy_secret_id('test_destroy_secret_id', secret_id=sec_id_result['secret_id'])

        with pytest.raises(aiovault.exceptions.InvalidPath):
            await client.auth.approle.lookup_secret_id('test_destroy_secret_id', secret_id=sec_id_result['secret_id'])

    async def test_approle_destroy_secret_id_accessor(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.approle.create('test_destroy_secret_id_accessor')
        sec_id_result = await client.auth.approle.generate_secret_id('test_destroy_secret_id_accessor', metadata={'test4': 'value4'})

        await client.auth.approle.destroy_secret_id('test_destroy_secret_id_accessor', secret_id_accessor=sec_id_result['secret_id_accessor'])

        with pytest.raises(aiovault.exceptions.InvalidPath):
            await client.auth.approle.lookup_secret_id('test_destroy_secret_id_accessor', secret_id_accessor=sec_id_result['secret_id_accessor'])

    async def test_login(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.approle.create('test_login')
        role_result = await client.auth.approle.read_roleid('test_login')
        assert 'role_id' in role_result

        secret_result = await client.auth.approle.generate_secret_id('test_login')
        assert 'secret_id' in secret_result

        login_result = await client.auth.approle.login(role_id=role_result['role_id'], secret_id=secret_result['secret_id'])
        assert 'client_token' in login_result.auth
        assert 'accessor' in login_result.auth


class TestLDAPAuth(BaseTestCase):
    @pytest.mark.run(order=1)
    async def test_mount_ldap(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        ldap_url = 'ldap://' + os.environ.get('TEST_LDAP_HOST', 'localhost') + ':' + os.environ.get('TEST_LDAP_PORT', '3389')  # Because my ldap server isnt on localhost
        await client.auth.ldap.mount('ldap', description='test_mount', url=ldap_url,
                                     discoverdn=False, userattr='uid', userdn='ou=people,dc=example,dc=com',
                                     groupdn='ou=groups,dc=example,dc=com', groupfilter='(member={{.UserDN}})', groupattr='cn')
        result = await client.auth.list_backends()
        assert 'ldap/' in result
        assert result['ldap/']['description'] == 'test_mount'

    async def test_config(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        ldap_backend = client.auth.get_ldap_backend('ldap')

        config = await ldap_backend.get_config()

        assert config['userdn'] == 'ou=people,dc=example,dc=com'
        assert config['groupdn'] == 'ou=groups,dc=example,dc=com'
        assert config['groupattr'] == 'cn'

    async def test_login(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        # Setup testgroup group to get default policy
        await client.auth.ldap.create_group('testgroup', policies=['default'])

        token_result = await client.auth.ldap.login('vault', 'test1234')

        assert 'client_token' in token_result.auth
        assert 'accessor' in token_result.auth
        assert 'default' in token_result.auth['policies']

    async def test_create_group(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.ldap.create_group('test_create_group', policies=['default'])

        result = await client.auth.ldap.list_groups()

        assert 'keys' in result
        assert 'test_create_group' in result['keys']

    async def test_read_group(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.ldap.create_group('test_read_group', policies=['default'])

        result = await client.auth.ldap.read_group('test_read_group')

        assert 'policies' in result
        assert result['policies'] == 'default'

    async def test_delete_group(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.ldap.create_group('test_delete_group', policies=['default'])
        result = await client.auth.ldap.list_groups()
        assert 'test_delete_group' in result['keys']

        await client.auth.ldap.delete_group('test_delete_group')
        result = await client.auth.ldap.list_groups()
        assert 'test_delete_group' not in result['keys']

    async def test_create_user(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.ldap.create_user('test_create_user', policies=['default'])

        result = await client.auth.ldap.list_users()

        assert 'keys' in result
        assert 'test_create_user' in result['keys']

    async def test_read_user(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.ldap.create_user('test_read_user', policies=['default'])

        result = await client.auth.ldap.read_user('test_read_user')

        assert 'policies' in result
        assert result['policies'] == 'default'

    async def test_delete_user(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.auth.ldap.create_user('test_delete_user', policies=['default'])
        result = await client.auth.ldap.list_users()
        assert 'test_delete_user' in result['keys']

        await client.auth.ldap.delete_user('test_delete_user')
        result = await client.auth.ldap.list_users()
        assert 'test_delete_user' not in result['keys']



# -*- coding: utf-8 -*-
"""Top-level package for aiovault."""

from .client import VaultClient  # noqa: F401

__author__ = """Terry Cain"""
__email__ = 'terry@terrys-home.co.uk'
__version__ = '0.1.0'

__all__ = ['client', 'exceptions']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `aiovault` package."""
import pytest
import asyncio
import aiovault
import aiovault.base

from .common import BaseTestCase, NoTLSTestCase


class TestBaseSys(BaseTestCase):
    async def test_remount(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.secrets.generic.mount(path='test_generic_rename')
        result = await client.secrets.list()
        assert 'test_generic_rename/' in result

        await client.sys.remount('test_generic_rename', 'test_generic_rename2')
        result = await client.secrets.list()
        assert 'test_generic_rename/' not in result
        assert 'test_generic_rename2/' in result


# Running these tests in order purely as it simplifies things
class TestSysInit(NoTLSTestCase):
    @pytest.mark.run(order=1)
    async def test_init(self, loop):
        await asyncio.sleep(2, loop=loop)

        client = aiovault.VaultClient(token=None, loop=loop)

        init_result = await client.sys.initialize(secret_shares=7, secret_threshold=4)

        assert 'root_token' in init_result
        assert 'keys' in init_result

        self.proc.root_token = init_result['root_token']
        self.proc.unseal_keys = init_result['keys']

    @pytest.mark.run(order=2)
    async def test_seal_status(self, loop):
        client = aiovault.VaultClient(token=None, loop=loop)

        result = await client.sys.seal_status()

        assert 'sealed' in result
        assert result['sealed']
        assert 'progress' in result
        assert 't' in result
        assert 'n' in result

    @pytest.mark.run(order=3)
    async def test_unseal(self, loop):
        client = aiovault.VaultClient(token=None, loop=loop)

        result = await client.sys.unseal(self.proc.unseal_keys[0])
        assert result['progress'] == 1
        result = await client.sys.unseal(self.proc.unseal_keys[1])
        assert result['progress'] == 2
        result = await client.sys.unseal(None)
        assert result['progress'] == 0

        await client.sys.unseal(self.proc.unseal_keys[0])  # 1
        await client.sys.unseal(self.proc.unseal_keys[1])  # 2
        await client.sys.unseal(self.proc.unseal_keys[2])  # 3
        result = await client.sys.unseal(self.proc.unseal_keys[3])  # 3

        assert not result['sealed']

    @pytest.mark.run(order=4)
    async def test_seal(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        result = await client.sys.seal_status()
        assert not result['sealed']

        await client.sys.seal()

        result = await client.sys.seal_status()
        assert result['sealed']

        await client.sys.unseal(self.proc.unseal_keys[0])  # 1
        await client.sys.unseal(self.proc.unseal_keys[1])  # 2
        await client.sys.unseal(self.proc.unseal_keys[2])  # 3
        result = await client.sys.unseal(self.proc.unseal_keys[3])  # 3

        assert not result['sealed']

    @pytest.mark.run(order=5)
    async def test_rotate(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        await client.sys.rotate()

    @pytest.mark.run(order=6)
    async def test_health(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        result = await client.sys.health()

        assert 'cluster_id' in result
        assert 'cluster_name' in result
        assert 'initialized' in result
        assert 'sealed' in result
        assert 'standby' in result
        assert 'version' in result

        assert not result['sealed']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `aiovault` package."""
import aiovault
import aiovault.base

from .common import BaseTestCase


class TestBasePolicy(BaseTestCase):
    async def test_list(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        result = await client.policies.list()
        assert 'policies' in result
        assert 'root' in result['policies']

    async def test_read(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        result = await client.policies.read('root')
        assert result['name'] == 'root'
        assert result['rules'] == ''

    async def test_create(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        policy = 'path "secret/*" { capabilities = ["read"] }'
        await client.policies.create('test_create', rules=policy)

        result = await client.policies.read('test_create')
        assert result['name'] == 'test_create'
        assert result['rules'] == policy

    async def test_update(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        policy1 = 'path "secret/*" { capabilities = ["read"] }'
        policy2 = 'path "secret/*" { capabilities = ["read", "list"] }'
        await client.policies.create('test_update', rules=policy1)
        await client.policies.update('test_update', rules=policy2)

        result = await client.policies.read('test_update')
        assert result['name'] == 'test_update'
        assert result['rules'] == policy2

    async def test_delete(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        policy = 'path "secret/*" { capabilities = ["read"] }'
        await client.policies.create('test_delete', rules=policy)
        await client.policies.delete('test_delete')



"""
Policy management
"""
from typing import Optional
from .base import HTTPBase, ResponseBase


class BasePolicy(HTTPBase):
    """
    Base policy class
    """
    async def create(self, name: str, rules: str):
        """
        Create policy

        :param name: Policy name
        :param rules: Policy HCL as a string
        """
        payload = {
            'rules': rules
        }

        await self._put(['sys/policy', name], payload=payload)

    async def update(self, name: str, rules: str):
        """
        Update policy (technically just overwrites it)

        :param name: Policy name
        :param rules: Policy HCL as a string
        """
        await self.create(name, rules)

    async def delete(self, name: str):
        """
        Delete policy

        :param name: Policy name
        """
        # can delete policy over and over again with a 200 OK
        await self._delete(['sys/policy', name])

    async def read(self, name: str, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        Read policy

        :param name: Policy name
        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._get(['sys/policy', name], wrap_ttl=wrap_ttl)
        json = await response.json()

        return ResponseBase(json_dict=json, request_func=self._request)

    async def list(self, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        List policies

        :param wrap_ttl: Wrap TTL
        :return: Response
        """
        response = await self._get('sys/policy', wrap_ttl=wrap_ttl)
        json = await response.json()

        return ResponseBase(json_dict=json, request_func=self._request)



"""
Vault Audit Backends
"""
from typing import Union, Optional
from .base import HTTPBase, ResponseBase


class FileBackend(HTTPBase):
    """
    Vault File Audit Backend
    """
    async def mount(self, mount_path: str, filepath: str, description: Optional[str] = None, log_raw: bool = False, hmac_accessor: bool = True,
                    mode: int = 0o0600, log_format: str = 'json', line_prefix: Optional[str] = ''):
        """
        Mount file audit backend

        :param mount_path: Mount path
        :param filepath: Log file
        :param description: Mount description
        :param log_raw: Log sensitive data without hashing
        :param hmac_accessor: Hash token accessor
        :param mode: Log file mode
        :param log_format: Log format, either json or jsonx
        :param line_prefix: Prefix log line with string
        """
        if log_format not in ('json', 'jsonx'):
            raise ValueError('log_format must be either "json" or "jsonx"')

        payload = {
            'type': 'file',
            'options': {
                'path': filepath,
                'log_raw': str(log_raw).lower(),
                'hmac_accessor': str(hmac_accessor).lower(),
                'mode': '{0:0>4o}'.format(mode),  # oct(0o0600) would generate '0o0600', vault wants 0600
                'format': log_format,
                'prefix': line_prefix

            }
        }
        if description is not None and isinstance(description, str):
            payload['description'] = description

        await self._put(['sys/audit', mount_path], payload=payload)


class SyslogBackend(HTTPBase):
    """
    Vault Syslog Audit Backend
    """
    async def mount(self, mount_path: str, description: Optional[str] = None, log_raw: bool = False, hmac_accessor: bool = True, log_format: str = 'json',
                    line_prefix: Optional[str] = '', facility: str = 'AUTH', tag: str = 'vault'):
        """
        Mount syslog audit backend

        :param mount_path: Mount path
        :param description: Mount description
        :param log_raw: Log sensitive data without hashing
        :param hmac_accessor: Hash token accessor
        :param log_format: Log format, either json or jsonx
        :param line_prefix: Prefix log line with string
        :param facility: Syslog facility
        :param tag: Syslog tag
        """
        if log_format not in ('json', 'jsonx'):
            raise ValueError('log_format must be either "json" or "jsonx"')

        payload = {
            'type': 'syslog',
            'options': {
                'facility': facility,
                'tag': tag,
                'log_raw': str(log_raw).lower(),
                'hmac_accessor': str(hmac_accessor).lower(),
                'format': log_format,
                'prefix': line_prefix

            }
        }
        if description is not None and isinstance(description, str):
            payload['description'] = description

        await self._put(['sys/audit', mount_path], payload=payload)


class BaseAudit(HTTPBase):
    """
    Base audit module.

    This class is sorta like a placeholder.
    It allows you to mount file backends like

    .. code-block:: python

        with aiovault.VaultClient(token='6c84fb90-12c4-11e1-840d-7b25c5ee775a') as client:
            client.audit.file.mount(...)
            # Or
            client.audit.file.syslog(...)
    """
    def __init__(self, *args, **kwargs):
        super(BaseAudit, self).__init__(*args, **kwargs)
        self._args = args
        self._kwargs = kwargs

        self._file = None
        self._syslog = None

    @property
    def file(self) -> FileBackend:
        """
        Get default file audit backend

        :return: File audit backend
        """
        if self._file is None:
            self._file = FileBackend(*self._args, **self._kwargs)

        return self._file

    @property
    def syslog(self) -> SyslogBackend:
        """
        Get default syslog audit backend

        :return: Syslog audit backend
        """
        if self._syslog is None:
            self._syslog = SyslogBackend(*self._args, **self._kwargs)

        return self._syslog

    async def list(self, wrap_ttl: Union[int, None] = None) -> ResponseBase:
        """
        List audit backends

        :param wrap_ttl: Wrap TTL
        :raises exceptions.VaultError: On error
        """
        response = await self._get('sys/audit', wrap_ttl=wrap_ttl)
        json = await response.json()

        return ResponseBase(json_dict=json, request_func=self._request)

    async def delete(self, path: str):
        """
        Delete an audit backend

        :param path: Audit backend mount path
        :raises exceptions.VaultError: On error
        """
        await self._delete(['sys/audit', path])

    async def hash(self, path: str, msg: str, wrap_ttl: Optional[int] = None) -> ResponseBase:
        """
        Get a HMAC of the input based on the given mount path

        .. code-block: python

            with aiovault.VaultClient(token='6c84fb90-12c4-11e1-840d-7b25c5ee775a') as client:
                result = await client.audit.hash('syslog', 'test1')
                print(result['hash'])  # hmac-sha256:SoMeHaShLaLaLaLa


        :param path: Audit backend mount path
        :param msg: Input to be hashed
        :param wrap_ttl: Wrap TTL
        :return: A response containing a key 'hash'
        :raises exceptions.VaultError: On error
        """
        payload = {'input': msg}
        response = await self._post(['sys/audit-hash', path], payload=payload, wrap_ttl=wrap_ttl)
        json = await response.json()

        return ResponseBase(json_dict=json, request_func=self._request)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `aiovault` package."""
import aiovault
import aiovault.base

from .common import BaseTestCase


class TestBasic(BaseTestCase):
    async def test_token_auth_success(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        result = await client.is_authenticated()

        assert result

    async def test_token_auth_failure(self, loop):
        client = aiovault.VaultClient(token=None, loop=loop)

        result = await client.is_authenticated()

        assert not result

    async def test_logout(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        result = await client.is_authenticated()
        assert result

        client.logout()
        assert client._auth_token is None

        result = await client.is_authenticated()
        assert not result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = ''.join(history_file.readlines()[1:])


requirements = [
    'aiohttp',
    'python-dateutil'
]

setup_requirements = [
    'pytest-runner'
]

test_requirements = [
    'pytest',
    'Sphinx',
    'sphinx-autodoc-typehints',
    'pytest-asyncio',
    'pytest-aiohttp',
    'requests',
    'coverage'
]

setup(
    name='aiovault',
    version='1.1.0',
    description="Vault asyncio",
    long_description=readme + '\n\n' + history,
    author="Terry Cain",
    author_email='terry@terrys-home.co.uk',
    url='https://github.com/terrycain/aiovault',
    packages=find_packages(include=['aiovault']),
    include_package_data=True,
    install_requires=requirements,
    license="GNU General Public License v3",
    zip_safe=False,
    keywords='aiovault',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `aiovault` package."""
import os
import pytest
import aiovault
import aiovault.base

from .common import BaseTestCase


class TestBaseAudit(BaseTestCase):
    async def test_backend_file_add(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        name = 'fileaudit'
        desc = 'Unittest Add'
        path = '/tmp/vaultaudit_unittest.log'

        await client.audit.file.mount(
            mount_path=name,
            filepath=path,
            description=desc,
        )

        audit_collection = await client.audit.list()

        assert isinstance(audit_collection, aiovault.base.ResponseBase)
        assert name + '/' in audit_collection

        file_mount = audit_collection[name + '/']
        assert file_mount['description'] == desc
        assert file_mount['options']['path'] == path

    @pytest.mark.skipif("TRAVIS" in os.environ, reason="Skipping this test on Travis CI.")
    async def test_backend_syslog_add(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        name = 'syslogaudit'
        desc = 'Unittest Add'
        tag = 'vaultsyslog'

        await client.audit.syslog.mount(mount_path=name, description=desc, tag=tag)

        audit_collection = await client.audit.list()

        assert isinstance(audit_collection, aiovault.base.ResponseBase)
        assert name + '/' in audit_collection

        file_mount = audit_collection[name + '/']
        assert file_mount['description'] == desc
        assert file_mount['options']['tag'] == tag

    async def test_backend_del(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        name = 'fileaudit_del'
        await client.audit.file.mount(mount_path=name, filepath='/tmp/vaultaudit_unittest_del.log', description='Unittest Del')

        audit_collection = await client.audit.list()
        assert name + '/' in audit_collection

        await client.audit.delete(name + '/')

        audit_collection = await client.audit.list()
        assert name + '/' not in audit_collection

    async def test_backend_hash(self, loop):
        client = aiovault.VaultClient(token=self.proc.root_token, loop=loop)

        name = 'fileaudit_hash'
        await client.audit.file.mount(mount_path=name, filepath='/tmp/vaultaudit_unittest_hash.log', description='Unittest Hash')

        audit_collection = await client.audit.list()
        assert name + '/' in audit_collection

        result = await client.audit.hash(name, 'test1')

        assert 'hash' in result
        assert result['hash'].startswith('hmac-sha256')
